from elfen.util import (
    rescale_column,
    normalize_column,
    filter_lexicon,
    upos_to_wn
)
import polars as pl
import pytest

@pytest.fixture
def sample_data():
    """
    Fixture to provide sample data for testing.
    """
    data = {
        'feature': [
            1.0, 2.0, 3.0, 4.0, 5.0,
            6.0, 7.0, 8.0, 9.0, 10.0
        ]
    }
    df = pl.DataFrame(data)
    return df

def test_rescale_column(sample_data):
    """
    Test the rescale_column function.
    """
    data = rescale_column(sample_data, 'feature', 0.0, 1.0)
    assert 'feature' in data.columns
    assert data['feature'].min() == 0.0
    assert data['feature'].max() == 1.0

def test_rescale_column_custom_range(sample_data):
    """
    Test the rescale_column function with a custom range.
    """
    data = rescale_column(sample_data, 'feature', 1.0, 5.0)
    assert 'feature' in data.columns
    assert data['feature'].min() == 1.0
    assert data['feature'].max() == 5.0

def test_rescale_column_no_change(sample_data):
    """
    Test the rescale_column function with no change.
    """
    data = rescale_column(sample_data, 'feature', 1.0, 10.0)
    assert 'feature' in data.columns
    assert data['feature'].min() == 1.0
    assert data['feature'].max() == 10.0
    for i in range(len(data)):
        assert data['feature'][i] == pytest.approx(sample_data['feature'][i])

def test_rescale_column_to_constant(sample_data):
    """
    Test the rescale_column function to a constant value.
    """
    data = rescale_column(sample_data, 'feature', 0.0, 0.0)
    assert 'feature' in data.columns
    assert data['feature'].min() == 0.0
    assert data['feature'].max() == 0.0
    for i in range(len(data)):
        assert data['feature'][i] == 0.0




import os

PROJECT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))

RESOURCE_MAP = {
    # Hedges, https://github.com/words/hedges
    "hedges": {
        "link": "https://raw.githubusercontent.com/words/hedges/main/data.txt",
        "area": "Semantics",
        "subarea": "Hedges",
        "filename": "hedges.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Semantics",
                                 "Hedges", "hedges.txt"),
        "bibtex": rb"""@misc{hedges, author = {Titus Wormer}, title = {Hedges}, year = {2022}, publisher = {GitHub}, journal = {GitHub repository}, howpublished = {\url{github.com/words/hedges}}, }"""
    },
    # Warriner, A. B., Kuperman, V., & Brysbaert, M. (2013).
    # Norms of valence, arousal, and dominance for 13,915 English lemmas.
    # Behavior Research Methods, 45(4), 1191-1207.
    "vad_warriner": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-012-0314-x/MediaObjects/"
                "13428_2012_314_MOESM1_ESM.zip",
        "area": "Emotion",
        "subarea": "VAD",
        "filename": "BRM-emot-submit.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "VAD", "BRM-emot-submit.csv"),
        "bibtex": rb"""
        @article{vad_warriner,
        title={Norms of valence, arousal, and dominance for 13,915 English lemmas},
        author={Warriner, Amy Beth and Kuperman, Victor and Brysbaert, Marc},
        journal={Behavior research methods},
        volume={45},
        pages={1191--1207},
        year={2013},
        publisher={Springer}
        }
        """
    },
    # Mohammad, S. M. (2018).
    # Obtaining reliable human ratings of valence, arousal, and dominance for 
    # 20,000 English words.
    # In Proceedings of the 56th Annual Meeting of the Association for
    # Computational Linguistics (Volume 1: Long Papers) (pp. 174-184).
    "vad_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/"
                "NRC-VAD-Lexicon.zip",
        "area": "Emotion",
        "subarea": "VAD",
        "filename": "NRC-VAD-Lexicon/NRC-VAD-Lexicon.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "VAD", "NRC-VAD-Lexicon", 
                                 "NRC-VAD-Lexicon.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources", 
                                              "Emotion", "VAD",
                                              "NRC-VAD-Lexicon",
                                              "NRC-VAD-Lexicon-"
                                              "ForVariousLanguages.txt"),
        "bibtex": rb"""
        @inproceedings{vad_nrc,
        title = "Obtaining Reliable Human Ratings of Valence, Arousal, and Dominance for 20,000 {E}nglish Words",
        author = "Mohammad, Saif",
        editor = "Gurevych, Iryna  and
        Miyao, Yusuke",
        booktitle = "Proceedings of the 56th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
        month = jul,
        year = "2018",
        address = "Melbourne, Australia",
        publisher = "Association for Computational Linguistics",
        url = "https://aclanthology.org/P18-1017",
        doi = "10.18653/v1/P18-1017",
        pages = "174--184",
        }
        """
    },
    # Mohammad, S. M. (2018)
    # Word affect intensity.
    # In Proceedings of the 56th Annual Meeting of the Association for
    # Computational Linguistics (Volume 1: Long Papers) (pp. 1609-1619).
    "intensity_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/"
                "NRC-Emotion-Intensity-Lexicon.zip",
        "area": "Emotion",
        "subarea": "Intensity",
        "filename": "NRC-Emotion-Intensity-Lexicon/"
                    "NRC-Emotion-Intensity-Lexicon-v1.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Intensity", "NRC-Emotion-Intensity-Lexicon",
                                 "NRC-Emotion-Intensity-Lexicon-v1.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                              "Emotion", "Intensity",
                                              "NRC-Emotion-Intensity-Lexicon/",
                                              "NRC-Emotion-Intensity-Lexicon"
                                              "-v1-ForVariousLanguages.txt"),
                                              
        "bibtex": rb"""
        @inproceedings{vad_nrc,
            title = "Word Affect Intensities",
            author = "Mohammad, Saif",
            editor = "Calzolari, Nicoletta  and
            Choukri, Khalid  and
            Cieri, Christopher  and
            Declerck, Thierry  and
            Goggi, Sara  and
            Hasida, Koiti  and
            Isahara, Hitoshi  and
            Maegaard, Bente  and
            Mariani, Joseph  and
            Mazo, H{\'e}l{\`e}ne  and
            Moreno, Asuncion  and
            Odijk, Jan  and
            Piperidis, Stelios  and
            Tokunaga, Takenobu",
            booktitle = "Proceedings of the Eleventh International Conference on Language Resources and Evaluation ({LREC} 2018)",
            month = may,
            year = "2018",
            address = "Miyazaki, Japan",
            publisher = "European Language Resources Association (ELRA)",
            url = "https://aclanthology.org/L18-1027",
        }
        """
    },
    # Brysbaert, M., Warriner, A. B., & Kuperman, V. (2014).
    # Concreteness ratings for 40 thousand generally known English word lemmas.
    # Behavior Research Methods, 46(3), 904-911.
    "concreteness_brysbaert": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-013-0403-5/MediaObjects/"
                "13428_2013_403_MOESM1_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "Concreteness",
        "filename": "13428_2013_403_MOESM1_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Concreteness", "13428_2013_403_MOESM1_ESM.xlsx"),
        "bibtex": rb"""
        @article{concreteness_brysbaert,
        title={Concreteness ratings for 40 thousand generally known English word lemmas},
        author={Brysbaert, Marc and Warriner, Amy Beth and Kuperman, Victor},
        journal={Behavior research methods},
        volume={46},
        pages={904--911},
        year={2014},
        publisher={Springer}
        }
        """
    },
    # Brysbaert, M., Mandera, P., McCormick, S. F., & Keuleers, E. (2019).
    # Word prevalence norms for 62,000 English lemmas.
    # Behavior Research Methods, 51(2), 467-479.
    "prevalence_brysbaert": {
        "link": "https://static-content.springer.com/"
                "esm/art%3A10.3758%2Fs13428-018-1077-9/"
                "MediaObjects/13428_2018_1077_MOESM2_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "Prevalence",
        "filename": "13428_2018_1077_MOESM2_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Prevalence", "13428_2018_1077_MOESM2_ESM.xlsx"),
        "bibtex": rb"""
        @article{prevalence_brysbaert,
        title={Word prevalence norms for 62,000 English lemmas},
        author={Brysbaert, Marc and Mandera, Pawe{\l} and McCormick, Samantha F and Keuleers, Emmanuel},
        journal={Behavior research methods},
        volume={51},
        pages={467--479},
        year={2019},
        publisher={Springer}
        }
        """
    },
    # Kuperman, V., Stadthagen-Gonzalez, H., & Brysbaert, M. (2013).
    # Age-of-acquisition ratings for 30,000 English words.
    # Behavior Research Methods, 45(4), 1191-1207.
    "aoa_kuperman": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-013-0348-8/"
                "MediaObjects/13428_2013_348_MOESM1_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "AgeOfAcquisition",
        "filename": "13428_2013_348_MOESM1_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "AgeOfAcquisition", "13428_2013_348_MOESM1_ESM.xlsx"),
        "bibtex": rb"""
        @article{aoa_kuperman,
        title={Age-of-acquisition ratings for 30,000 English words},
        author={Kuperman, Victor and Stadthagen-Gonzalez, Hans and Brysbaert, Marc},
        journal={Behavior research methods},
        volume={44},
        pages={978--990},
        year={2012},
        publisher={Springer}
        }
        """
    },
    # Baccianella, S., Esuli, A., & Sebastiani, F. (2010).
    # SentiWordNet 3.0: An enhanced lexical resource for sentiment
    # analysis and opinion mining.
    # In LREC (Vol. 10, pp. 2200-2204).
    "sentiwordnet": { # Currently not supported
        "link": "https://raw.githubusercontent.com/aesuli/SentiWordNet/"
                "master/data/SentiWordNet_3.0.0.txt",
        "area": "Emotion",
        "subarea": "Sentiment",
        "filename": "SentiWordNet_3.0.0.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Sentiment", "SentiWordNet_3.0.0.txt"),
        "bibtex": rb""  # TODO
    },
        # Mohammad, S. M., & Turney, P. D. (2013).
    # Emotions evoked by common words and phrases:
    # Using Mechanical Turk to create an emotion lexicon.
    # In Proceedings of the NAACL HLT 2013 Workshop on
    # Computational Approaches to Analysis and Generation
    # of Emotion in Text (pp. 26-34).
    "sentiment_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/NRC-Emotion-Lexicon.zip",
        "area": "Emotion",
        "subarea": "Sentiment",
        "filename": "NRC-Emotion-Lexicon/NRC-Emotion-Lexicon-Wordlevel-v0.92.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Sentiment", "NRC-Emotion-Lexicon",
                                 "NRC-Emotion-Lexicon-Wordlevel-v0.92.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                              "Emotion", "Sentiment",
                                              "NRC-Emotion-Lexicon",
                                              "NRC-Emotion-Lexicon-"
                                              "ForVariousLanguages.txt"),
        "bibtex": rb"""
        @article{sentiment_nrc,
        Author = {Mohammad, Saif M. and Turney, Peter D.},
        Journal = {Computational Intelligence},
        Number = {3},
        Pages = {436--465},
        Title = {Crowdsourcing a Word-Emotion Association Lexicon},
        Volume = {29},
        Year = {2013}
        }
        """
    },
    # Coso, B., Guasch, M., Buganovic, I., Ferre, P., & Hinojosa, J. A. (2022).
    # CROWD-5e: A croatian psycholinguistic database for affective norms for
    # five discrete emotions.
    # Behavior Research Methods, 55(1), 4018-4034.
    # TODO: Handle processing the data
    "affect_crowd5e": {
        "link": "https://figshare.com/ndownloader/files/36434421",
        "area": "Emotion",
        "subarea": "Affect",
        "filename": "CROWD-5e.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Affect", "CROWD-5e.xlsx"),
        "bibtex": rb""  # TODO
    },
    # Diveica, V, Pexman, P. M., & Binney, R. J. (2023).
    # Quantifying Social Semantics: An Inclusive Definition
    # of Socialness and Ratings for 8,388 English Words.
    # Behavior Research Methods, 55, 461-473.
    "socialness": {
        "link": "https://osf.io/download/29eyh/",
        "area": "Psycholinguistics",
        "subarea": "Socialness",
        "filename": "Socialness.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Socialness", "Socialness.csv"),
        "bibtex": rb"""
        @article{socialness,
        title={Quantifying social semantics: An inclusive definition of socialness and ratings for 8388 English words},
        author={Diveica, Veronica and Pexman, Penny M and Binney, Richard J},
        journal={Behavior Research Methods},
        volume={55},
        number={2},
        pages={461--473},
        year={2023},
        publisher={Springer}
        }
        """
    },
     # Lynott, D., Connell, L., Brysbaert, M., Brand, J., & Carney, J. (2020).
    # The Lancaster Sensorimotor Norms: Multidimensional measures of
    # perceptual and action strength for 40,000 English words.
    # Behavior Research Methods, 52, 1271-1291.
    "sensorimotor_lancaster": {
        "link": "https://osf.io/download/pu85v/",
        "area": "Psycholinguistics",
        "subarea": "Sensorimotor",
        "filename": "LancasterSensorimotorNorms.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Sensorimotor", "LancasterSensorimotorNorms.xlsx"),
        "bibtex": rb"""
        @article{sensorimotor_lancaster,
        title={The Lancaster Sensorimotor Norms: multidimensional measures of perceptual and action strength for 40,000 English words},
        author={Lynott, Dermot and Connell, Louise and Brysbaert, Marc and Brand, James and Carney, James},
        journal={Behavior research methods},
        volume={52},
        pages={1271--1291},
        year={2020},
        publisher={Springer}
        }
        """
    },
    # Winter, B., Lupyan, G., Perry, L. K., Dingemanse, M., & Perlman, M. (2021).
    # Iconicity ratings for 14,000 English words.
    # Behavior Research Methods, 56, 1640-1655.
    "iconicity_winter": {
        "link": "https://osf.io/download/ex37k/",
        "area": "Psycholinguistics",
        "subarea": "Iconicity",
        "filename": "WinterIconicityNorms.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics",
                                 "Iconicity", "WinterIconicityNorms.csv"),
        "bibtex": rb"""
        @article{iconicity_winter,
        title={Iconicity ratings for 14,000+ English words},
        author={Winter, Bodo and Lupyan, Gary and Perry, Lynn K and Dingemanse, Mark and Perlman, Marcus},
        journal={Behavior research methods},
        volume={56},
        number={3},
        pages={1640--1655},
        year={2024},
        publisher={Springer}
        }
        """
    },
    # Bonin, P., M\'eot, A., & Burgiska, A. (2018).
    # Concreteness ratings for 1,659 French words.
    # Behavior Research Methods, 50(6), 2366-2387.
    # has concreteness, aoa, and other ratings
    "concreteness_bonin": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-018-1014-y/MediaObjects/"
                "13428_2018_1014_MOESM3_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "Concreteness",
        "filename": "BoninConcretenessNorms.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics",
                                 "Concreteness",
                                 "BoninConcretenessNorms.xlsx"),
        "bibtex": rb"""
        @article{concreteness_bonin,
        title={Concreteness ratings for 1,659 French words: Relationships with other psycholinguistic variables and word recognition times},
        author={Bonin, Patrick and M\'eot, Aur\'elie and Burgiska, Aur\'elia},
        journal={Behavior Research Methods},
        volume={50},
        pages={2366--2387},
        year={2018},
        publisher={Springer}
        }
        """
    },
    # AoA German
    "aoa_schroeder": {
        "link": "https://static-content.springer.com/"
                "esm/art%3A10.3758%2Fs13428-011-0164-y/"
                "MediaObjects/13428_2011_164_MOESM1_ESM.xls",
        "area": "Psycholinguistics",
        "subarea": "AgeofAcquisition",
        "filename": "SchroederAoANorms.xls",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics",
                                 "AgeofAcquisition", "SchroederAoANorms.xls"),
        "bibtex": rb"""
        @article{aoa_schroeder,
        title={German norms for semantic typicality, age of acquisition, and concept familiarity},
        author={Schroeder, Sabine and Gemballa, Teresa and Ruppin, Steffie and Wartenburger, Isabelle},
        journal={Behavior Research Methods},
        volume={44},
        pages={380--394},
        year={2011},
        publisher={Springer}
        }
        """
    },
    # concreteness German
    "leipzig_affective_norms": {
        "link": "https://static-content.springer.com/"
                "esm/art%3A10.3758%2FBRM.42.4.987/"
                "MediaObjects/Kanske-BRM-2010.zip",
        "area": "Psycholinguistics",
        "subarea": "Concreteness",
        "filename": "Kanske-BRM-2010/LANG_database.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics",
                                 "Concreteness", 
                                 "Kanske-BRM-2010",
                                 "LANG_database.txt"),
        "bibtex": rb"""
        @article{leipzig_affective_norms,
        title={The Leipzig Affective Norms for German: A reliability study},
        author={Kanske, Philipp and Kotz, Sonja A.},
        journal={Behavior Research Methods},
        volume={42},
        number={4},
        pages={987--991},
        year={2010},
        publisher={Springer}
        }
        """
    },
    # sensorimotor norms for italian
    "sensorimotor_vergallito": {
        "link": "https://osf.io/download/qhywk/",
        "area": "Psycholinguistics",
        "subarea": "Sensorimotor",
        "filename": "vergallito_sensorimotor.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics", "Sensorimotor",
                                 "vergallito_sensorimotor.txt"),
        "bibtex": rb"""
        @article{sensorimotor_vergallito,
        title={Perceptual modality norms for 1,121 Itlaian words: A comparison with concreteness and imageability scores and an analysis of their impact on word processing tasks},
        author={Vergallito, Alessandra and Petilli, Marco Alessandro and Marelli, Marco},
        journal={Behavior Research Methods},
        volume={52},
        pages={1599--1614},
        year={2020},
        publisher={Springer}
        }
        """
    },
    # concreteness norms for italian
    "affective_norms_montefinese": {
        "link": "https://osf.io/download/cdu3b/",
        "area": "Psycholinguistics",
        "subarea": "Concreteness",
        "filename": "2014-Montefinese_Database.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics", "Concreteness",
                                 "2014-Montefinese_Database.xlsx"),
        "bibtex": rb"""
        @article{affective_norms_montefinese,
        title={The adaptation of the Affective Norms for English Words (ANEW) in Italian},
        author={Montefinese, Maria and Ambrosini, Elena and Fairfield, Beth and Mammarella, Nicola},
        journal={Behavior Research Methods},
        volume={46},
        pages={887--903},
        year={2014},
        publisher={Springer}
        """
    },
    # age of acquisition norms for italian
    "aoa_montefinese": {
        "link": "https://osf.io/download/3nvh6/",
        "area": "Psycholinguistics",
        "subarea": "AgeofAcquisition",
        "filename": "ItAoA.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                 "Psycholinguistics", "Age of Acquisition",
                                 "ItAoA.xlsx"),
        "bibtex": rb"""
        @article{aoa_montefinese,
        AUTHOR={Montefinese, Maria  and Vinson, David  and Vigliocco, Gabriella  and Ambrosini, Ettore }, 
        TITLE={Italian Age of Acquisition Norms for a Large Set of Words (ItAoA)},
        JOURNAL={Frontiers in Psychology},
        VOLUME={Volume 10 - 2019},
        YEAR={2019},
        DOI={10.3389/fpsyg.2019.00278}
        }
        """
    },
    # "perceptual_chedid": {
    # # TODO: Has two sheets, one for visual and one for auditory
    # # TBD how to handle this
    #     # https://lingualab.ca/en/project/norms-familiarity-perceptual-strength/
    #     "link": "",
    #     "area": "Psycholinguistics",
    #     "subarea": "Sensorimotor",
    #     "filename": "ChedidPerceptualNorms.xlsx",
    #     "filepath": os.path.join(PROJECT_PATH, "elfen_resources",
    #                              "Psycholinguistics", "Sensorimotor",
    #                              "ChedidPerceptualNorms.xlsx"),
    #     "bibtex": rb"""
    #     @article{perceptual_chedid,
    #     title={Visual and auditory perceptual strength norms for 3,596 French nouns and their relationship with other psycholinguistic variables},
    #     author={Chedid, Georges and Brambati, Simona Maria and Bedetti, Christophe and Rey, Amandine E. and Wilson, Maximiliano A. and Vallet, Guillaume T.},
    #     journal={Behavior Research Methods},
    #     volume={51},
    #     pages={2094--2105},
    #     year={2019},
    #     publisher={Springer}
    #     }
    #     """
    # }
}




"""
This module contains functions to download external resources.

If you are using the resources for research, please cite the original
authors.
"""
import requests
import zipfile
import os

from .resource_utils.langs import LANGUAGES_NRC
from .resource_utils.resource_map import RESOURCE_MAP

PROJECT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


def download_lexicon(link: str,
                     path: str,
                     filename: str = None
                     ) -> None:
    """
    Download a lexicon from a link and save it to a path.
    
    Args:
        link (str): Link to the lexicon.
        path (str): Path to save the lexicon.
        filename (str): Name of the file to save the lexicon.

    Returns:
        None
    """
    # Headers to avoid 406 response
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64;"
                      " rv:91.0) Gecko/20100101 Firefox/91.0",
        "Accept": "text/html,application/xhtml+xml,application/"
                  "xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
        "Pragma": "no-cache",
        "Cache-Control": "no-cache"
    }
    response = requests.get(link, headers=headers)

    if filename is None:
        filename = link.split("/")[-1]

    if link.endswith(".zip"):
        with open("temp.zip", "wb") as f:
            f.write(response.content)
        with zipfile.ZipFile("temp.zip", "r") as zip_ref:
            zip_ref.extractall(path)
        os.remove("temp.zip")
    elif link.endswith(".xlsx"):
        # filename = link.split("/")[-1]
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    elif link.endswith(".txt"):
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    else:
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    
def get_resource(feature: str) -> None:
    """
    Download a resource from the RESOURCE_MAP.

    Args:
        feature (str): Name of the feature to download.

    Returns:
        None
    """
    if feature not in RESOURCE_MAP:
        raise ValueError(f"Feature {feature} not found in RESOURCE_MAP.")

    # Making sure all the necessary directories exist
    os.makedirs(os.path.join(PROJECT_PATH, "elfen_resources",
                             RESOURCE_MAP[feature]["area"],
                             RESOURCE_MAP[feature]["subarea"]),
                             exist_ok=True)
    # Downloading the lexicon if it does not exist
    if not os.path.exists(RESOURCE_MAP[feature]["filepath"]):
        download_lexicon(RESOURCE_MAP[feature]["link"],
                         os.path.join(PROJECT_PATH, "elfen_resources",
                                      RESOURCE_MAP[feature]["area"],
                                      RESOURCE_MAP[feature]["subarea"]),
                                      RESOURCE_MAP[feature]["filename"])

def list_external_resources() -> None:
    """
    List all the external resources available in the RESOURCE_MAP.

    Args:
        None

    Returns:
        None
    """
    for feature in RESOURCE_MAP:
        print(f"Feature: {feature}")
        print("\n")

def get_bibtex() -> str:
    """
    Print the bibtex citation for all the resources in the RESOURCE_MAP,
    and the package itself.

    The citation keys for the lexicons are the names of the lexicons in
    the RESOURCE_MAP.
    The citation key for the package is "maurer-2024-elfen".

    Args:
        None

    Returns:
        bibxtex (str): Bibtex citation for all the resources.
    """

    bibtex = rb"""
    @misc{maurer-2025-elfen,
        author = {Maurer, Maximilian},
        title = {ELFEN - Efficient Linguistic Feature Extraction for Natural Language Datasets},
        year = {2025},
        publisher = {GitHub},
        journal = {GitHub repository},
        howpublished = {\url{https://github.com/mmmaurer/elfen}},
    }
    """
    for feature in RESOURCE_MAP:
        bibtex += RESOURCE_MAP[feature]["bibtex"]

    return bibtex.decode("utf-8")




SENSORIMOTOR_VARS = {
    "en" : [
        "Auditory",
        "Gustatory",
        "Haptic",
        "Interoceptive",
        "Olfactory",
        "Visual",
        "Foot_leg",
        "Hand_arm",
        "Head",
        "Mouth",
        "Torso",
    ],
    "it" : [
        "Auditory",
        "Gustatory",
        "Haptic",
        "Olfactory",
        "Visual"
    ]
}


"""
This module contains functions to extract named entity-related features
from text data.
"""
import polars as pl


ENT_TYPES = [
    'ORG', 'CARDINAL', 'DATE', 'GPE', 'PERSON', 'MONEY', 'PRODUCT', 'TIME',
    'PERCENT', 'WORK_OF_ART', 'QUANTITY', 'NORP', 'LOC', 'EVENT', 'ORDINAL',
    'FAC', 'LAW', 'LANGUAGE'
]

def get_num_entities(data: pl.DataFrame,
                     backbone: str = 'spacy',
                     **kwargs: dict[str, str],
                     ) -> pl.DataFrame:
    """
    Calculates the number of entities in the text data.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of entities in the
            text data.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(x.ents),
                                       return_dtype=pl.UInt16
                                       ).alias("n_entities"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(x.entities),
                                       return_dtype=pl.UInt16
                                       ).alias("n_entities"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
    
    return data

def get_num_per_entity_type(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            ent_types: list = ENT_TYPES,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of entities per entity type in the text data.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        ent_types: A list of entity types to calculate the number of
                entities for. Default is the list of entity types in the
                spaCy/stanza libraries.
    
    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the number of entities per 
            entity type in the text data.
    """
    if backbone == 'spacy':
        for ent_type in ent_types:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: len(
                    [ent for ent in x.ents if ent.label_ == ent_type]),
                    return_dtype=pl.UInt16
                    ).alias(f"n_{ent_type.lower()}"),
            )
    elif backbone == 'stanza':
        for ent_type in ent_types:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: len(
                    [ent for ent in x.entities if ent.type == ent_type]),
                    return_dtype=pl.UInt16
                    ).alias(f"n_{ent_type.lower()}"),
            )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
    
    return data




"""
This module contains helper mappings to map feature names to the
respective functions that calculate them. Additionally, it contains
mappings to group features by their respective feature areas.
"""
from .preprocess import (
    get_lemmas,
    get_tokens,
)
from .dependency import (
    # tree features
    get_tree_width,
    get_tree_depth,
    get_tree_branching,
    get_ramification_factor,
    # dependency features
    get_n_noun_chunks,
    get_n_per_dependency_type
)

from .emotion import (
    # sentiment
    get_sentiment_score,
    get_n_negative_sentiment,
    get_n_positive_sentiment,
    # emotion intensity
    get_avg_emotion_intensity,
    get_n_low_intensity,
    get_n_high_intensity,
    # VAD
    get_avg_valence,
    get_avg_arousal,
    get_avg_dominance,
    get_n_low_valence,
    get_n_high_valence,
    get_n_low_arousal,
    get_n_high_arousal,
    get_n_low_dominance,
    get_n_high_dominance,
)
from .entities import (
    get_num_entities,
    get_num_per_entity_type,
)
from .information import (
    get_compressibility,
    get_entropy,
)
from .lexical_richness import (
    get_lemma_token_ratio,
    get_ttr,
    get_cttr,
    get_rttr,
    get_herdan_c,
    get_summer_index,
    get_dugast_u,
    get_maas_index,
    get_n_hapax_legomena,
    get_n_global_lemma_hapax_legomena,
    get_n_global_token_hapax_legomena,
    get_n_hapax_dislegomena,
    get_n_global_lemma_hapax_dislegomena,
    get_n_global_token_hapax_dislegomena,
    get_lexical_density,
    get_hdd,
    get_sichel_s,
    get_global_sichel_s,
    get_giroud_index,
    get_mtld,
    get_mattr,
    get_msttr,
    get_yule_k,
    get_simpsons_d,
    get_herdan_v,
)
from .morphological import (
    get_morph_feats,
)
from .pos import (
    get_num_lexical_tokens,
    get_pos_variability,
    get_num_per_pos,
)
from .psycholinguistic import (
    # Concreteness
    get_avg_concreteness,
    get_n_high_concreteness,
    get_n_low_concreteness,
    get_avg_sd_concreteness,
    get_n_controversial_concreteness,
    # AoA
    get_avg_aoa,
    get_n_high_aoa,
    get_n_low_aoa,
    get_avg_sd_aoa,
    get_n_controversial_aoa,
    # Prevalence
    get_avg_prevalence,
    get_n_high_prevalence,
    get_n_low_prevalence,
    # Socialness
    get_avg_socialness,
    get_n_high_socialness,
    get_n_low_socialness,
    get_avg_sd_socialness,
    get_n_controversial_socialness,
    # Iconicity
    get_avg_iconicity,
    get_n_high_iconicity,
    get_n_low_iconicity,
    get_avg_sd_iconicity,
    get_n_controversial_iconicity,
    # Sensorimotor
    get_avg_sensorimotor,
    get_n_high_sensorimotor,
    get_n_low_sensorimotor,
    get_avg_sd_sensorimotor,
    get_n_controversial_sensorimotor,
)
from .readability import (
    get_flesch_reading_ease,
    get_flesch_kincaid_grade,
    get_smog,
    get_ari,
    get_cli,
    get_gunning_fog,
    get_lix,
    get_rix,
    get_num_monosyllables,
    get_num_polysyllables,
    get_num_syllables,
)
from .resource_utils.langs import LANGUAGES_NRC
from .semantic import (
    get_num_hedges,
    get_avg_num_synsets,
    get_avg_num_synsets_per_pos,
    get_num_high_synsets,
    get_num_low_synsets,
    get_high_synsets_per_pos,
    get_low_synsets_per_pos,
)
from .surface import (
    get_avg_word_length,
    get_num_lemmas,
    get_num_long_words,
    get_num_sentences,
    get_num_tokens,
    get_num_tokens_per_sentence,
    get_num_types,
    get_raw_sequence_length,
    get_num_characters
)

FUNCTION_MAP = {
    # SURFACE FEATURES
    "tokens": get_tokens,
    "lemmas": get_lemmas,
    "raw_sequence_length": get_raw_sequence_length,
    "n_tokens": get_num_tokens,
    "n_lemmas": get_num_lemmas,
    "n_sentences": get_num_sentences,
    "n_types": get_num_types,
    "avg_word_length": get_avg_word_length,
    "n_long_words": get_num_long_words,
    "n_tokens_per_sentence": get_num_tokens_per_sentence,
    "n_characters": get_num_characters,
    # MORPHOLOGICAL FEATURES
    "n_per_morph_feature": get_morph_feats,
    # EMOTION FEATURES
    # Sentiment features
    "sentiment_score": get_sentiment_score,
    "n_positive_sentiment": get_n_positive_sentiment,
    "n_negative_sentiment": get_n_negative_sentiment,
    # VAD features
    "avg_valence": get_avg_valence,
    "avg_arousal": get_avg_arousal,
    "avg_dominance": get_avg_dominance,
    "n_low_valence": get_n_low_valence,
    "n_high_valence": get_n_high_valence,
    "n_low_arousal": get_n_low_arousal,
    "n_high_arousal": get_n_high_arousal,
    "n_low_dominance": get_n_low_dominance,
    "n_high_dominance": get_n_high_dominance,
    # Emotion intensity features
    "avg_emotion_intensity": get_avg_emotion_intensity,
    "n_low_intensity": get_n_low_intensity,
    "n_high_intensity": get_n_high_intensity,
    # ENTITY FEATURES
    "n_entities": get_num_entities,
    "n_per_entity_type": get_num_per_entity_type,
    # INFORMATION THEORETIC FEATURES
    "compressibility": get_compressibility,
    "entropy": get_entropy,
    # LEXICAL RICHNESS FEATURES
    "lemma_token_ratio": get_lemma_token_ratio,
    "ttr": get_ttr,
    "cttr": get_cttr,
    "rttr": get_rttr,
    "herdan_c": get_herdan_c,
    "summer_index": get_summer_index,
    "dugast_u": get_dugast_u,
    "maas_index": get_maas_index,
    "n_hapax_legomena": get_n_hapax_legomena,
    "n_global_lemma_hapax_legomena": get_n_global_lemma_hapax_legomena,
    "n_global_token_hapax_legomena": get_n_global_token_hapax_legomena,
    "lexical_density": get_lexical_density,
    "n_hapax_dislegomena": get_n_hapax_dislegomena,
    "n_global_lemma_hapax_dislegomena": get_n_global_lemma_hapax_dislegomena,
    "n_global_token_hapax_dislegomena": get_n_global_token_hapax_dislegomena,
    "hdd": get_hdd,
    "sichel_s": get_sichel_s,
    "global_sichel_s": get_global_sichel_s,
    "giroud_index": get_giroud_index,
    "mtld": get_mtld,
    "mattr": get_mattr,
    "msttr": get_msttr,
    "yule_k": get_yule_k,
    "simpsons_d": get_simpsons_d,
    "herdan_v": get_herdan_v,
    # POS FEATURES
    "n_lexical_tokens": get_num_lexical_tokens,
    "pos_variability": get_pos_variability,
    "n_per_pos": get_num_per_pos,
    # PSYCHOLINGUISTIC FEATURES
    # Concreteness
    "avg_concreteness": get_avg_concreteness,
    "n_high_concreteness": get_n_high_concreteness,
    "n_low_concreteness": get_n_low_concreteness,
    "avg_sd_concreteness": get_avg_sd_concreteness,
    "n_controversial_concreteness": get_n_controversial_concreteness,
    # AoA
    "avg_aoa": get_avg_aoa,
    "n_high_aoa": get_n_high_aoa,
    "n_low_aoa": get_n_low_aoa,
    "avg_sd_aoa": get_avg_sd_aoa,
    "n_controversial_aoa": get_n_controversial_aoa,
    # Prevalence
    "avg_prevalence": get_avg_prevalence,
    "n_high_prevalence": get_n_high_prevalence,
    "n_low_prevalence": get_n_low_prevalence,
    # Socialness
    "avg_socialness": get_avg_socialness,
    "n_high_socialness": get_n_high_socialness,
    "n_low_socialness": get_n_low_socialness,
    "avg_sd_socialness": get_avg_sd_socialness,
    "n_controversial_socialness": get_n_controversial_socialness,
    # Iconicity
    "avg_iconicity": get_avg_iconicity,
    "n_high_iconicity": get_n_high_iconicity,
    "n_low_iconicity": get_n_low_iconicity,
    "avg_sd_iconicity": get_avg_sd_iconicity,
    "n_controversial_iconicity": get_n_controversial_iconicity,
    # Sensorimotor
    "avg_sensorimotor": get_avg_sensorimotor,
    "n_high_sensorimotor": get_n_high_sensorimotor,
    "n_low_sensorimotor": get_n_low_sensorimotor,
    "avg_sd_sensorimotor": get_avg_sd_sensorimotor,
    "n_controversial_sensorimotor": get_n_controversial_sensorimotor,
    # READABILITY FEATURES
    "flesch_reading_ease": get_flesch_reading_ease,
    "flesch_kincaid_grade": get_flesch_kincaid_grade,
    "smog": get_smog,
    "ari": get_ari,
    "cli": get_cli,
    "gunning_fog": get_gunning_fog,
    "lix": get_lix,
    "rix": get_rix,
    "n_monosyllables": get_num_monosyllables,
    "n_polysyllables": get_num_polysyllables,
    "n_syllables": get_num_syllables,
    # SEMANTIC FEATURES
    "n_hedges": get_num_hedges,
    "avg_num_synsets": get_avg_num_synsets,
    "avg_num_synsets_per_pos": get_avg_num_synsets_per_pos,
    "n_high_synsets": get_num_high_synsets,
    "n_low_synsets": get_num_low_synsets,
    "n_high_synsets_per_pos": get_high_synsets_per_pos,
    "n_low_synsets_per_pos": get_low_synsets_per_pos,
    # DEPENDENCY FEATURES
    # tree features
    "tree_width": get_tree_width,
    "tree_depth": get_tree_depth,
    "tree_branching": get_tree_branching,
    "ramification_factor": get_ramification_factor,
    # dependency features
    "n_noun_chunks": get_n_noun_chunks,
    "n_per_dependency_type": get_n_per_dependency_type,
}

FEATURE_AREA_MAP = {
    "surface": [
        "raw_sequence_length",
        "n_tokens",
        "n_lemmas",
        "n_sentences",
        "n_types",
        "avg_word_length",
        "n_long_words",
        "n_tokens_per_sentence",
        "n_characters",
    ],
    "morphological": [
        "n_per_morph_feature",
    ],
    "dependency": [
        "tree_width",
        "tree_depth",
        "tree_branching",
        "ramification_factor",
        "n_noun_chunks",
        "n_per_dependency_type",
    ],
    "emotion": [
        "sentiment_score",
        "n_positive_sentiment",
        "n_negative_sentiment",
        "avg_valence",
        "avg_arousal",
        "avg_dominance",
        "n_low_valence",
        "n_high_valence",
        "n_low_arousal",
        "n_high_arousal",
        "n_low_dominance",
        "n_high_dominance",
        "avg_emotion_intensity",
        "n_low_intensity",
        "n_high_intensity",
    ],
    "entities": [
        "n_entities",
        "n_per_entity_type",
    ],
    "information": [
        "compressibility",
        "entropy",
    ],
    "lexical_richness": [
        "lemma_token_ratio",
        "ttr",
        "cttr",
        "rttr",
        "herdan_c",
        "summer_index",
        "dugast_u",
        "maas_index",
        "n_hapax_legomena",
        "n_global_lemma_hapax_legomena",
        "n_global_token_hapax_legomena",
        "lexical_density",
        "n_hapax_dislegomena",
        "n_global_lemma_hapax_dislegomena",
        "n_global_token_hapax_dislegomena",
        "hdd",
        "sichel_s",
        "global_sichel_s",
        "giroud_index",
        "mtld",
        "mattr",
        "msttr",
        "yule_k",
        "simpsons_d",
        "herdan_v",
    ],
    "pos": [
        "n_lexical_tokens",
        "pos_variability",
        "n_per_pos",
    ],
    "psycholinguistic": [
        "avg_concreteness",
        "n_high_concreteness",
        "n_low_concreteness",
        "avg_sd_concreteness",
        "n_controversial_concreteness",
        "avg_aoa",
        "n_high_aoa",
        "n_low_aoa",
        "avg_sd_aoa",
        "n_controversial_aoa",
        "avg_prevalence",
        "n_high_prevalence",
        "n_low_prevalence",
        "avg_socialness",
        "n_high_socialness",
        "n_low_socialness",
        "avg_sd_socialness",
        "n_controversial_socialness",
        "avg_iconicity",
        "n_high_iconicity",
        "n_low_iconicity",
        "avg_sd_iconicity",
        "n_controversial_iconicity",
        "avg_sensorimotor",
        "n_high_sensorimotor",
        "n_low_sensorimotor",
        "avg_sd_sensorimotor",
        "n_controversial_sensorimotor",
    ],
    "readability": [
        "flesch_reading_ease",
        "flesch_kincaid_grade",
        "smog",
        "ari",
        "cli",
        "gunning_fog",
        "lix",
        "rix",
        "n_monosyllables",
        "n_polysyllables",
        "n_syllables",
    ],
    "semantic": [
        "n_hedges",
        "avg_num_synsets",
        "avg_num_synsets_per_pos",
        "n_high_synsets",
        "n_low_synsets",
        "n_high_synsets_per_pos",
        "n_low_synsets_per_pos",
    ],
}

# TODO: transform into language-specific mappings
FEATURE_LEXICON_MAP = {
    "n_hedges": {"en": "hedges"},
    # nrc lexicons are multilingual
    "sentiment_score": {**{key: "sentiment_nrc" for key in LANGUAGES_NRC},
                        **{"en": "sentiment_nrc"}},
    "n_positive_sentiment": {**{key: "sentiment_nrc" for key in LANGUAGES_NRC},
                             **{"en": "sentiment_nrc"}},
    "n_negative_sentiment": {**{key: "sentiment_nrc" for key in LANGUAGES_NRC},
                             **{"en": "sentiment_nrc"}},
    "avg_valence": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                    **{"en": "vad_nrc"}},
    "avg_arousal": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                    **{"en": "vad_nrc"}},
    "avg_dominance": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                      **{"en": "vad_nrc"}},
    "n_low_valence": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                      **{"en": "vad_nrc"}},
    "n_high_valence": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                       **{"en": "vad_nrc"}},
    "n_low_arousal": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                       **{"en": "vad_nrc"}},
    "n_high_arousal": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                       **{"en": "vad_nrc"}},
    "n_low_dominance": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                        **{"en": "vad_nrc"}},
    "n_high_dominance": {**{key: "vad_nrc" for key in LANGUAGES_NRC},
                         **{"en": "vad_nrc"}},
    "avg_emotion_intensity": {**{key: "intensity_nrc" for key in LANGUAGES_NRC},
                              **{"en": "intensity_nrc"}},
    "n_low_intensity": {**{key: "intensity_nrc" for key in LANGUAGES_NRC},
                         **{"en": "intensity_nrc"}},
    "n_high_intensity": {**{key: "intensity_nrc" for key in LANGUAGES_NRC},
                         **{"en": "intensity_nrc"}},
    # language-specific psycholinguistic features
    "avg_concreteness": {"en": "concreteness_brysbaert",
                         "fr": "concreteness_bonin",
                         "de": "leipzig_affective_norms",
                         "it": "affective_norms_montefinese"},
    "n_high_concreteness": {"en": "concreteness_brysbaert",
                            "fr": "concreteness_bonin",
                            "de": "leipzig_affective_norms",
                            "it": "affective_norms_montefinese"},
    "n_low_concreteness": {"en": "concreteness_brysbaert",
                           "fr": "concreteness_bonin",
                           "de": "leipzig_affective_norms",
                           "it": "affective_norms_montefinese"},
    "avg_sd_concreteness": {"en": "concreteness_brysbaert",
                            "fr": "concreteness_bonin",
                            "de": "leipzig_affective_norms",
                            "it": "affective_norms_montefinese"},
    "n_controversial_concreteness": {"en": "concreteness_brysbaert",
                                     "fr": "concreteness_bonin",
                                     "de": "leipzig_affective_norms",
                                     "it": "affective_norms_montefinese"},
    "avg_aoa": {"en": "aoa_kuperman",
                "de": "aoa_schroeder",
                "it": "aoa_montefinese"},
    "n_high_aoa": {"en": "aoa_kuperman",
                   "de": "aoa_schroeder",
                   "it": "aoa_montefinese"},
    "n_low_aoa": {"en": "aoa_kuperman",
                  "de": "aoa_schroeder",
                  "it": "aoa_montefinese"},
    "avg_sd_aoa": {"en": "aoa_kuperman",
                   "de": "aoa_schroeder",
                   "it": "aoa_montefinese"},
    "n_controversial_aoa": {"en": "aoa_kuperman",
                             "de": "aoa_schroeder",
                             "it": "aoa_montefinese"},
    "avg_prevalence": {"en": "prevalence_brysbaert"},
    "n_high_prevalence": {"en": "prevalence_brysbaert"},
    "n_low_prevalence": {"en": "prevalence_brysbaert"},
    "avg_socialness": {"en": "socialness"},
    "n_high_socialness": {"en": "socialness"},
    "n_low_socialness": {"en": "socialness"},
    "avg_sd_socialness": {"en": "socialness"},
    "n_controversial_socialness": {"en": "socialness"},
    "avg_iconicity": {"en": "iconicity_winter"},
    "n_high_iconicity": {"en": "iconicity_winter"},
    "n_low_iconicity": {"en": "iconicity_winter"},
    "avg_sd_iconicity": {"en": "iconicity_winter"},
    "n_controversial_iconicity": {"en": "iconicity_winter"},
    "avg_sensorimotor": {"en": "sensorimotor_lancaster",
                         "it": "sensorimotor_vergallito"},
    "n_high_sensorimotor": {"en": "sensorimotor_lancaster",
                            "it": "sensorimotor_vergallito"},
    "n_low_sensorimotor": {"en": "sensorimotor_lancaster",
                           "it": "sensorimotor_vergallito"},
    "avg_sd_sensorimotor": {"en": "sensorimotor_lancaster",
                            "it": "sensorimotor_vergallito"},
    "n_controversial_sensorimotor": {"en": "sensorimotor_lancaster",
                                     "it": "sensorimotor_vergallito"},
}




import pytest
import polars as pl
from elfen import Extractor

@pytest.fixture
def sample_data_en():
    """
    Fixture to provide sample data for testing.
    """
    data = {
        'text': [
            "This is a test sentence.",
            "Another test sentence.",
            "Yet another test sentence."
        ]
    }
    df = pl.DataFrame(data)
    return df

@pytest.fixture
def sample_data_de():
    """
    Fixture to provide sample data for testing in German.
    """
    data = {
        'text': [
            "Dies ist ein Testsatz.",
            "Noch ein Testsatz.",
            "Wieder ein weiterer Testsatz."
        ]
    }
    df = pl.DataFrame(data)
    return df

def test_initialization(sample_data_en):
    """
    Test the initialization of the Extractor class.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    assert extractor.data is not None
    assert 'nlp' in extractor.data.columns

def test_extraction_single_feature(sample_data_en):
    """
    Test the extraction of features from the text data.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    extractor.extract("n_tokens")
    assert 'n_tokens' in extractor.data.columns
    assert len(extractor.data['n_tokens']) == len(sample_data_en)
    assert extractor.data.shape == (3, 3)  # 3 rows, 3 columns (text, nlp, n_tokens)

def test_extraction_multiple_features(sample_data_en):
    """
    Test the extraction of multiple features from the text data.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    extractor.extract(["n_tokens", "n_sentences"])
    assert 'n_tokens' in extractor.data.columns
    assert 'n_sentences' in extractor.data.columns
    assert len(extractor.data['n_tokens']) == len(sample_data_en)
    assert len(extractor.data['n_sentences']) == len(sample_data_en)
    assert extractor.data.shape == (3, 4)  # 3 rows, 4 columns (text, nlp, n_tokens, n_sentences)

def test_extraction_invalid_feature(sample_data_en):
    """
    Test the extraction of an invalid feature from the text data.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    extractor.extract("invalid_feature")
    assert 'invalid_feature' not in extractor.data.columns

def test_initialization_invalid_backbone(sample_data_en):
    """
    Test the extraction with an invalid backbone.
    """
    with pytest.raises(ValueError):
        Extractor(data=sample_data_en,
                  backbone='invalid_backbone',
                  text_column='text',
                  language='en',
                  model='en_core_web_sm')

def test_token_normalization(sample_data_en):
    """
    Test the token normalization feature.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    extractor.extract(["n_tokens", "n_long_words"])
    extractor.token_normalize()
    first_ratio = extractor.data['n_long_words'][0]
    max_tokens = extractor.data['n_tokens'].max()
    # make sure the ratio works as expected
    assert pytest.approx(first_ratio) == 1/6 
    # make sure n_tokens is not normalized
    assert max_tokens == 6

def test_empty_text_warning():
    """
    Test whether a warning is raised for empty texts.
    """
    data = {
        'text': [
            "This is a test sentence.",
            "",
            "Yet another test sentence."
        ]
    }
    df = pl.DataFrame(data)
    with pytest.warns(UserWarning, match="Some texts are empty. "
                      "This can affect the results. "
                      "You may want to remove these rows."):
        Extractor(data=df,
                  backbone='spacy',
                  text_column='text',
                  language='en',
                  model='en_core_web_sm')

def test_full_run_en(sample_data_en):
    """
    Test whether the Extractor runs without errors for English texts.
    """
    extractor = Extractor(data=sample_data_en,
                          backbone='spacy',
                          text_column='text',
                          language='en',
                          model='en_core_web_sm')
    extractor.extract_features()
    assert extractor.data is not None

def test_full_run_de(sample_data_de):
    """
    Test whether the Extractor runs without errors for German texts.
    """
    extractor = Extractor(data=sample_data_de,
                          backbone='spacy',
                          text_column='text',
                          language='de',
                          model='de_core_news_sm')
    extractor.extract_features()
    assert extractor.data is not None




"""
This module contains utility functions for working with Polars DataFrames.
"""
import warnings
import polars as pl

def rescale_column(data: pl.DataFrame,
                   column: str,
                   minimum: float = 0.0,
                   maximum: float = 1.0) -> pl.DataFrame:
    """
    Rescales a column to a custom range. Defaults to [0, 1].

    Args:
        data (pl.DataFrame): A Polars DataFrame.
        column (str): The name of the column to rescale.
        minimum (float): 
            The desired minimum value of the column.
            Defaults to 0.
        maximum (float):
            The desired maximum value of the column.
            Defaults to 1.

    Returns:
        rescaled_data (pl.DataFrame):
            A Polars DataFrame with the column rescaled to
            [minimum, maximum].
    """
    inner_minimum = data[column].min()
    inner_maximum = data[column].max()
    
    # Rescale to [0, 1]
    if minimum == 0.0 and maximum == 1.0:
        rescaled_data = data.with_columns([
            ((pl.col(column) - inner_minimum) / 
            (inner_maximum - inner_minimum)).alias(column)
        ])
    # Rescale to [minimum, maximum]
    else:
        rescaled_data = data.with_columns([
            (((pl.col(column) - inner_minimum) / 
            (inner_maximum - inner_minimum)) * 
            (maximum - minimum) + minimum).alias(column)
        ])

    return rescaled_data

def normalize_column(data: pl.DataFrame,
                     column: str) -> pl.DataFrame:
    """
    Normalizes a column to have a mean of 0 and a standard deviation of 1.

    Args:
        data (pl.DataFrame): A Polars DataFrame.
        column (str): The name of the column to normalize.
    
    Returns:
        normalized_data (pl.DataFrame):
            A Polars DataFrame with the column normalized
    """
    mean = data[column].mean()
    std = data[column].std()

    normalized_data = data.with_columns([
        ((pl.col(column) - mean) / std).alias(column)
    ])

    return normalized_data

def filter_lexicon(lexicon: pl.DataFrame,
                   words: pl.Series,
                   word_column: str = "Word"
                   ) -> pl.DataFrame:
    """
    Filters a lexicon to only include the words in a list.

    Args:
        lexicon (pl.DataFrame): A Polars DataFrame containing the lexicon.
        words (pl.Series): A Polars Series containing the words to include.
        word_column (str):
            The name of the column containing the words in the lexicon.

    Returns:
        filtered_lexicon (pl.DataFrame):
            A Polars DataFrame containing only the words in the list.
    """
    return lexicon.filter(pl.col(word_column).is_in(words))

def upos_to_wn(upos_tag: str) -> str:
    """
    Converts a Universal POS tag to a (Senti)WordNet POS tag.

    Args:
        upos_tag (str): A Universal POS tag.

    Returns:
        wn_tag (str): A WordNet POS tag.
    """
    if upos_tag in {"NOUN", "PROPN"}:
        return "n"
    elif upos_tag in {"VERB", "AUX"}:
        return "v"
    elif upos_tag in {"ADV"}:
        return "r"
    elif upos_tag in {"ADJ"}:
        return "a"
    else:
        return None
    
def zero_token_warning_nan(feature: str) -> None:
    """
    Generates a warning message for features that cannot be calculated
    for texts with zero tokens. Warning is to be issued if any NaN values
    are present in the feature.

    Args:
        feature (str): The name of the feature.

    Returns:
        None
    """
    warnings.warn(f"Some texts have 0 tokens, resulting in NaN "
                  f"values for the {feature}. Consider filtering "
                  f"these texts or filling NaN values using "
                  f"`fill_nan()` or removing them for further analysis.")

def zero_token_warning_null(feature: str) -> None:
    """
    Generates a warning message for features that cannot be calculated
    for texts with zero tokens. Warning is to be issued if any Null values
    are present in the feature.

    Args:
        feature (str): The name of the feature.

    Returns:
        None
    """
    warnings.warn(f"Some texts have 0 tokens, resulting in Null "
                  f"values for the {feature}. Consider filtering "
                  f"these texts or filling Null values using "
                  f"`fill_null()` or removing them for further analysis.")
    


"""
This module contains the Extractor class. The Extractor class is the main
class in the ELFEN package and is used to extract features from text data.
"""
import os
import re
from typing import Union
import warnings

import polars as pl

from .configs.extractor_config import (
    CONFIG_ALL,
)
from .preprocess import (
    preprocess_data,
)
from .features import (
    FUNCTION_MAP,
    FEATURE_AREA_MAP,
    FEATURE_LEXICON_MAP,
)
from .emotion import (
    load_sentiment_nrc_lexicon,
    load_intensity_lexicon,
    load_vad_lexicon,
)
from .psycholinguistic import (
    load_concreteness_norms,
    load_aoa_norms,
    load_prevalence_norms,
    load_socialness_norms,
    load_sensorimotor_norms,
    load_iconicity_norms,
)
from .resources import (
    RESOURCE_MAP,
    get_resource,
)
from .ratios import (
    get_feature_token_ratio,
    get_feature_type_ratio,
    get_feature_sentence_ratio,
)
from .semantic import (
    load_hedges,
)

from .surface import (
    get_global_lemma_frequencies,
    get_global_token_frequencies,
    get_raw_sequence_length,
)

from .util import (
    rescale_column,
    normalize_column,
)


class Extractor:
    """
    The Extractor class is the main class in the ELFEN package and is used
    to extract features from text data.

    The Extractor class takes a Polars DataFrame containing text data and
    a configuration dictionary as input. The configuration dictionary
    specifies the text column, the NLP library to use as backbone, the
    language of the text data, the model to use for processing the text
    data, and the features to extract.

    The Extractor class has methods to extract features from the text data,
    extract features in a feature group, extract a single feature, get the
    names of the extracted features, write the extracted features to a CSV
    file, and clean up the extracted features.

    To access the data with the extracted features, use the data attribute
    of the Extractor class: ``extractor.data``.
    """
    def __init__(self, 
                 data: pl.DataFrame,
                 config: dict[str, str] = CONFIG_ALL,
                 **kwargs,
                 ) -> None:
        self.data = data
        self.config = config
        self.basic_features = []
        self.ratio_features = {
            "type": [],
            "token": [],
            "sentence": [],
        }

        # Handle typical keyword arguments
        if "language" in kwargs:
            self.config["language"] = kwargs["language"]
        if "backbone" in kwargs:
            self.config["backbone"] = kwargs["backbone"]
        if "model" in kwargs:
            self.config["model"] = kwargs["model"]
        if "text_column" in kwargs:
            self.config["text_column"] = kwargs["text_column"]
        if "max_length" in kwargs:
            self.config["max_length"] = kwargs["max_length"]
        if "remove_constant_cols" in kwargs:
            self.config["remove_constant_cols"] = kwargs[
                "remove_constant_cols"]

        if "max_length" in self.config:
            max_length = self.config["max_length"]
        else:
            max_length = 1_000_000

        # Check if the backbone is valid
        if self.config["backbone"] not in ["spacy", "stanza"]:
            raise ValueError("Backbone must be 'spacy' or 'stanza'.")

        self.data = preprocess_data(data=self.data,
                                    text_column=self.config["text_column"],
                                    backbone=self.config["backbone"],
                                    lang=self.config["language"],
                                    model=self.config["model"],
                                    max_length=max_length)
        
        self.helper_cols = [
            "nlp",
            "lemmas",
            "tokens",
            "token_freqs",
            'synsets',
            'synsets_noun',
            'synsets_verb',
            'synsets_adj',
            'synsets_adv',
        ]
        self.initial_cols = self.data.columns

        # Warning if there are empty texts
        if get_raw_sequence_length(self.data,
                                   text_column=self.config["text_column"]
                                   ).filter(
            pl.col("raw_sequence_length") == 0
        ).shape[0] > 0:
            warnings.warn("Some texts are empty. This can affect the "
                          "results. You may want to remove these rows.")

    def __apply_function(self,
                         feature,
                         function_map = FUNCTION_MAP,
                         **kwargs,
                         ) -> None:
        """
        Helper function to apply a feature extraction function to the
        data. Handles features that require additional parameters.

        NOTE: Currently works for all kwargs except pos_tags
        specification. This is a known issue that will be fixed in
        future versions.
        This is a minor problem though, as pos feature extraction
        is fast; removing the unwanted pos tag features is a bit
        of an inconvenience, but not a major issue.
        Major functionality like specifying lexicons, thresholds, etc. is
        supported.  

        Args:
            feature: The feature to extract.
            function_map: A dictionary of feature extraction functions.
            **kwargs: 
                Additional keyword arguments for the feature extraction.
                Any lexicons or thresholds required for the feature, or
                additional parameters the respective feature function
                takes.
        """
        backbone = self.config["backbone"]
        text_column = self.config["text_column"]
  
        self.data = function_map[feature](
            data=self.data,
            backbone=backbone,
            text_column=text_column,
            language=self.config["language"],
            **kwargs)

    def extract_feature_group(self,
                              feature_group: Union[str, list[str]],
                              feature_area_map: dict[str, str] = \
                                FEATURE_AREA_MAP
                             ) -> None:
        """
        Extract all features in a feature group with default settings.
        Available feature groups are dependency, emotion, entities,
        information, lexical_richness, morphological, pos,
        psycholinguistic, readability, semantic, and surface.

        NOTE: Multilingual support for features that require lexicons or
        norms is currently only implemented for emotion/sentiment features.
        For pscholinguistic features and hedges, only English resources
        are currently available.

        Args:
            feature_group (str): 
                The feature group to extract features from.
            feature_area_map (dict[str, str]):
                A dictionary mapping features to feature areas.

        Returns:
            data (pl.DataFrame):
                The data with the extracted features.
        """
        if type(feature_group) == str:
            feature_group = [feature_group]
        for group in feature_group:
            if group in feature_area_map:
                for feature in feature_area_map[group]:
                    print(f"Extracting {feature}...")
                    if feature in FEATURE_LEXICON_MAP:
                        lexicon = self.__gather_resource_from_featurename(
                            language=self.config["language"],
                            feature=feature,
                            feature_lexicon_map=FEATURE_LEXICON_MAP)
                        if lexicon is not None:
                            print(f"Extracting {feature}...")
                            self.__apply_function(feature,
                                                  lexicon=lexicon)
                    elif feature in FUNCTION_MAP:
                        self.__apply_function(feature)
                    else:
                        print(f"Feature {feature} not found. Check "
                              "spelling.")
            else:
                print("Feature group not found. Check spelling.")
    
    def __load_lexicon_from_featurename(self,
                                        filepath: str,
                                        featurename: str,
                                        ) -> pl.DataFrame:
        """
        Helper function to load lexicons from the resources.

        Args:
            filepath (str): The path to the lexicon.
            featurename (str): The name of the feature.

        Returns:
            lexicon (pl.DataFrame):
                The lexicon to use for feature extraction.
        """
        if "aoa" in featurename:
            lexicon = load_aoa_norms(filepath, 
                                     language=self.config["language"])
        elif "concreteness" in featurename:
            lexicon = load_concreteness_norms(filepath,
                                              language=self.config["language"])
        elif "prevalence" in featurename:
            lexicon = load_prevalence_norms(filepath,
                                            language=self.config["language"])
        elif re.search(r"(valence|arousal|dominance)",
                       featurename):
            lexicon = load_vad_lexicon(filepath,
                                       language=self.config["language"])
        elif "sentiment" in featurename:
            lexicon = load_sentiment_nrc_lexicon(filepath,
                                                 language=self.config[
                                                     "language"])
        elif "intensity" in featurename:
            lexicon = load_intensity_lexicon(filepath,
                                             language=self.config[
                                                 "language"])
        elif "hedges" in featurename:
            lexicon = load_hedges(filepath,
                                   language=self.config["language"])
        elif "socialness" in featurename:
            lexicon = load_socialness_norms(filepath,
                                             language=self.config["language"])
        elif "sensorimotor" in featurename:
            lexicon = load_sensorimotor_norms(filepath,
                                               language=self.config["language"])
        elif "iconicity" in featurename:
            lexicon = load_iconicity_norms(filepath,
                                            language=self.config["language"])
        else:
            print(f"Feature {featurename} not found. Skipping...")
            lexicon = None
        return lexicon
    
    def token_normalize(self,
                        features: Union[list[str], str] = "all",
                        **kwargs,
                        ) -> None:
        """
        Normalize the occurence-based features with the number of tokens.

        Args:
            features (Union[list[str], str]):
                The features to normalize. Default is "all".
                Allows for a list of features or a single feature in str
                format, or 'all' to normalize all occurence-based
                features.
        
        Returns:
            None
        """
        if type(features) == list:
            for feature in features:
                self.data = self.data.with_columns(
                    (pl.col(feature) / pl.col("n_tokens")). \
                        alias(feature)
                )
        elif features == "all":
            # exclude n_tokens, n_types, and n_sentences
            feats = [f for f in self.get_feature_names() if f not in
                        ["n_tokens",
                         "n_types",
                         "n_sentences",
                         "n_lemmas",
                         "n_syllables"] and
                        f.startswith("n_")]
            for feature in feats:
                self.data = self.data.with_columns(
                    (pl.col(feature) / pl.col("n_tokens")). \
                        alias(feature)
                )

    def ratio_normalize(self,
                        features: Union[list[str], str] = "all",
                        ratio: str = "type",
                        **kwargs,
                        ) -> None:
        """
        Normalize the extracted features with a specific ratio feature.
        Ratios available are type, token, and sentence ratios.

        Args:
            features (Union[list[str], str]):
                The features to normalize. Default is "all".
                Allows for a list of features or a single feature in str
                format, or 'all' to normalize all features.
            ratio (str):
                The ratio to normalize the features with.
                Default is "type".

        Returns:
            None
        """
        if ratio == "type":
            ratio_fct = get_feature_type_ratio
        elif ratio == "token":
            ratio_fct = get_feature_token_ratio
        elif ratio == "sentence":
            ratio_fct = get_feature_sentence_ratio
        else:
            print("Ratio not found. Check spelling.")
            return
        
        if type(features) == list:
            self.data = ratio_fct(data=self.data,
                                  features=features,
                                  backbone=self.config["backbone"])
        else:
            if features == "all":
                features = self.get_feature_names()
                self.data = ratio_fct(data=self.data,
                                      features=features,
                                      backbone=self.config["backbone"])
            else:  # single feature in str format
                self.data = ratio_fct(data=self.data,
                                      features=[features],
                                      backbone=self.config["backbone"])

    def normalize(self,
                  features: Union[list[str], str] = "all",
                  **kwargs,
                  ) -> None:
        """
        Normalize the extracted features to have a mean of 0 and a
        standard deviation of 1.

        Args:
            features (Union[list[str], str]):
                The features to normalize. Default is "all".
                Allows for a list of features or a single feature in str
                format, or 'all' to normalize all features.

        Returns:
            None
        """
        if type(features) == list:
            for feature in features:
                self.data = normalize_column(data=self.data,
                                             column=feature)
        else:
            if features == "all":
                features = self.get_feature_names()
                for feature in features:
                    self.data = normalize_column(data=self.data,
                                                 column=feature)
            else:
                self.data = normalize_column(data=self.data,
                                             column=features)
        
    def rescale(self,
                features: Union[list[str], str] = "all",
                minimum: float = 0,
                maximum: float = 1,
                **kwargs,
                ) -> None:
        """
        Rescale the extracted features to a specific range.

        Args:
            features (Union[list[str], str]):
                The features to rescale. Default is "all".
                Allows for a list of features or a single feature in str
                format, or 'all' to rescale all features.
            minimum (float):
                The minimum value to rescale the features to.
                Default is 0.
            maximum (float):
                The maximum value to rescale the features to.
                Default is 1.

        Returns:
            None
        """
        if type(features) == list:
            for feature in features:
                self.data = rescale_column(data=self.data,
                                           column=feature,
                                           minimum=minimum,
                                           maximum=maximum)
        else:
            if features == "all":
                features = self.get_feature_names()
                for feature in features:
                    self.data = rescale_column(data=self.data,
                                               column=feature,
                                               minimum=minimum,
                                               maximum=maximum)
            else:
                self.data = rescale_column(data=self.data,
                                           column=features,
                                           minimum=minimum,
                                           maximum=maximum)

    def __gather_resource_from_featurename(self,
                                           language: str,
                                           feature: str,
                                           feature_lexicon_map: 
                                           dict[str, str] = \
                                               FEATURE_LEXICON_MAP
                                           ) -> pl.DataFrame:
        """
        Helper function to gather resources for feature extraction.

        TODO:
            - Add handling for languages other than English.

        Args:
            feature (str): The feature to extract.
            feature_lexicon_map (dict[str, str]):
                A dictionary mapping features to lexicons.
        
        Returns:
            lexicon (pl.DataFrame):
                The lexicon to use for feature extraction.
        """
        if feature in feature_lexicon_map:
            if language in feature_lexicon_map[feature].keys():
                if feature_lexicon_map[feature][language] in  \
                    RESOURCE_MAP and "multilingual_filepath" not in \
                    RESOURCE_MAP[
                        feature_lexicon_map[feature][language]]:
                    filepath = RESOURCE_MAP[
                        feature_lexicon_map[feature][language]]["filepath"]
                elif "multilingual_filepath" in RESOURCE_MAP[
                    feature_lexicon_map[feature][language]]:
                    if language != "en":
                        filepath = RESOURCE_MAP[
                            feature_lexicon_map[feature][language]][
                                "multilingual_filepath"]
                    elif language == "en":
                        filepath = RESOURCE_MAP[
                            feature_lexicon_map[feature][language]][
                                "filepath"]
                else:
                    print(f"Feature {feature} not (yet) "
                          f"supported for {language}. "
                          "Skipping...")
                    return None
                if not os.path.exists(filepath):
                    get_resource(feature_lexicon_map[feature][language])
                lexicon = self.__load_lexicon_from_featurename(
                    filepath, feature)
                return lexicon
            else:
                print(f"Feature {feature} not (yet) supported for "
                      f"{language}. Skipping...")
                return None
        else:
            print(f"Feature {feature} not found. Skipping...")
            return None

    def extract_features(self) -> None:
        """
        Extracts all features specified in the config.

        Returns:
            None
        """
        features = self.config["features"]
        for feature_area in features:
            for feature in features[feature_area]:
                print(f"Extracting {feature}...")
                if feature in FUNCTION_MAP:
                    if feature in FEATURE_LEXICON_MAP:
                        lexicon = self.__gather_resource_from_featurename(
                            language=self.config["language"],
                            feature=feature,
                            feature_lexicon_map=FEATURE_LEXICON_MAP
                        )
                        if lexicon is not None:
                            self.__apply_function(feature,
                                                  lexicon=lexicon)
                    else:
                        self.__apply_function(feature)
                else:
                    print(f"Feature {feature} not found. Check spelling. "
                          "Skipping...")

        # Remove constant columns if specified and if there is more than 
        # one row
        if self.config["remove_constant_cols"] and len(self.data) > 1:
            self.remove_constant_cols()
    
    def extract(self,
                features: Union[str|list[str]],
                **kwargs,
                ):
        """
        Extract a single feature from the data.

        Args:
            features (Union[str, list[str]]):
                The feature to extract. Can be a single feature in str
                format or a list of features.
            **kwargs:
                Additional keyword arguments for the feature extraction.
                Any lexicons or thresholds required for the feature, or
                additional parameters the respective feature function
                takes.

        Returns:
            None
        """
        if type(features) == str:
            features = [features]
        for feature_name in features:
            if feature_name in FUNCTION_MAP:
                # handle lexicon separately from other kwargs to ensure
                # that the lexicon is loaded correctly if specified by 
                # the user
                if "lexicon" in kwargs:
                    lexicon = kwargs["lexicon"]
                    self.__apply_function(feature_name,
                                        **kwargs)
                elif feature_name in FEATURE_LEXICON_MAP and \
                    "lexicon" not in kwargs:
                    lexicon = self.__gather_resource_from_featurename(
                        language=self.config["language"],
                        feature=feature_name,
                        feature_lexicon_map=FEATURE_LEXICON_MAP)
                    if lexicon is not None:
                        self.__apply_function(feature_name,
                                              lexicon=lexicon,
                                              **kwargs)
                else:
                    self.__apply_function(feature_name)
            else:
                print(f"Feature {feature_name} not found. Check spelling.")
    
    def __cleanup_cols(self):
        """
        Helper function to remove helper columns from the data.
        """
        if self.config["remove_constant_cols"] and len(self.data) > 1:
            self.remove_constant_cols()
        self.data = self.data.drop(
            set(self.data.columns).intersection(set(self.helper_cols)))

    def remove_constant_cols(self):
        """
        Helper function to remove constant columns from the data.
        Constant columns are columns with only one unique value.
        """
        # Remove constant feature columns
        cols_to_drop = [col for col in self.get_feature_names() if
                        self.data[col].n_unique() == 1]
        
        self.data = self.data.drop(cols_to_drop)

    def get_feature_names(self) -> list[str]:
        """
        Get the names of the extracted features.

        Returns:
            feature_names (list[str]):
                A list of the names of the extracted features.
        """
        return list(
            set(self.data.columns) - set(self.initial_cols + \
                                            self.helper_cols))
    
    def write_csv(self,
                  filepath: str,
                  **kwargs,
                  ) -> None:
        """
        Save the extracted features to a CSV file.
        Cleans up the data by removing helper columns before saving and
        then writes the data to a CSV using Polars DataFrame to_csv.
        To specify additional keyword arguments for the to_csv method,
        pass them as keyword arguments to this function.

        Args:
            filepath (str): The path to save the CSV file.
            **kwargs:
                Additional keyword arguments for the Polars DataFrame
                to_csv method.
        """
        self.__cleanup_cols()
        self.data.write_csv(filepath, **kwargs)

    def get_data(self) -> pl.DataFrame:
        """
        Get the data with the extracted features.

        Returns:
            data (pl.DataFrame):
                The data with the extracted features.
        """
        return self.data
    
    def get_corpus_frequencies(self,
                               kind: str = "token",
                               ) -> dict[str, int]:
        """
        Get the global frequencies of tokens or lemmas in the data.

        Args:
            kind (str): 
                The kind of frequency to get. Default is "token".
                Options are "token" or "lemma".
        
        Returns:
            frequencies (dict[str, int]):
                A dictionary of the global frequencies of tokens or lemmas
                in the data.
        """
        if kind == "token":
            frequencies = get_global_token_frequencies(
                data=self.data,
                backbone=self.config["backbone"])
        elif kind == "lemma":
            frequencies = get_global_lemma_frequencies(
                data=self.data,
                backbone=self.config["backbone"])
            
        return frequencies




"""
This module contains functions to calculate readability scores from text
data. Readability scores are used to assess the readability, i.e., the
complexity, of a text.

The readability scores implemented in this module are:

- Total number of syllables
- Number of monosyllables
- Number of polysyllables
- Flesch Reading Ease
- Flesch-Kincaid Grade Level
- Automated Readability Index (ARI)
- Simple Measure of Gobbledygook (SMOG)
- Coleman-Liau Index (CLI)
- Gunning Fog Index
- LIX
- RIX
"""
import polars as pl

from .surface import (
    get_num_tokens,
    get_num_sentences,
    get_num_characters,
    get_num_long_words
)
from .util import (
    zero_token_warning_nan
)

def get_num_syllables(data: pl.DataFrame,
                      backbone: str = 'spacy',
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
        """
        Calculates the number of syllables in a text.

        Syllables are calculated using the syllables_count attribute
        of the tokens in the text data.

        Args:
            data (pl.DataFrame):
                A Polars DataFrame containing the text data.
            backbone (str):
                The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
                Not supported for Stanza backbone.

        Returns:
            data (pl.DataFrame):
                A Polars DataFrame containing the number of syllables
                in the text data.
                The number of syllables is stored in a new column
                named 'n_syllables'.
        """
        if backbone == 'spacy':
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: sum(
                    [token._.syllables_count for token in x if 
                     token._.syllables_count != None]),
                     return_dtype=pl.UInt16
                    ).alias("n_syllables"),
            )
        elif backbone == 'stanza':
            raise NotImplementedError(
                "Not supported for Stanza backbone."
            )
        else:
            raise ValueError(f"Unsupported backbone '{backbone}'. "
                             "Supported backbones are 'spacy' and 'stanza'.")
        
        return data

def get_num_monosyllables(data: pl.DataFrame,
                          backbone: str = 'spacy',
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the number of monosyllables in a text.

    Monosyllables are words with one syllable.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of monosyllables
            in the text data.
            The number of monosyllables is stored in a new column
            named 'n_monosyllables'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: sum(
                [1 for syllables_count in [
                    token._.syllables_count
                    for token
                    in x
                    if token._.syllables_count != None
                    ] if syllables_count == 1]
                    ), return_dtype=pl.UInt16
                ).alias("n_monosyllables"),
        )
    elif backbone == 'stanza':
        raise NotImplementedError(
            "Not supported for Stanza backbone."
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")

    return data

def get_num_polysyllables(data: pl.DataFrame,
                          backbone: str = 'spacy',
                          **kwargs: dict[str, str],
                          ) -> pl.DataFrame:
    """
    Calculates the number of polysyllables in a text.

    Polysyllables are words with three or more syllables.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of polysyllables
            in the text data.
            The number of polysyllables is stored in a new column
            named 'n_polysyllables'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: sum(
                [1 for syllables_count in [
                    token._.syllables_count
                    for token
                    in x
                    if token._.syllables_count != None
                    ] if syllables_count >= 3]
                    ), return_dtype=pl.UInt16
                ).alias("n_polysyllables"),
        )
    elif backbone == 'stanza':
        raise NotImplementedError(
            "Not supported for Stanza backbone."
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")

    return data

def get_flesch_reading_ease(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the Flesch Reading Ease score of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Flesch Reading Ease score
            of the text data. The Flesch Reading Ease score is stored
            in a new column named 'flesch_reading_ease'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_syllables' not in data.columns:
        data = get_num_syllables(data, backbone=backbone)

    data = data.with_columns(
        (206.835 - (1.015 * (pl.col("n_tokens") / \
                             pl.col("n_sentences"))) - \
         (84.6 * (pl.col("n_syllables") / pl.col("n_sentences")))
         ).alias("flesch_reading_ease"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("flesch_reading_ease")

    return data

def get_flesch_kincaid_grade(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the Flesch-Kincaid Grade Level of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Flesch-Kincaid Grade Level
            of the text data. The Flesch-Kincaid Grade Level is stored
            in a new column named 'flesch_kincaid_grade'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_syllables' not in data.columns:
        data = get_num_syllables(data, backbone=backbone)

    data = data.with_columns(
        (0.39 * (pl.col("n_tokens") / pl.col("n_sentences")) + \
         11.8 * (pl.col("n_syllables") / pl.col("n_sentences")) - 15.59
         ).alias("flesch_kincaid_grade"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("flesch_kincaid_grade")

    return data

def get_ari(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the Automated Readability Index (ARI) of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Automated Readability Index
            of the text data. The Automated Readability Index is stored
            in a new column named 'ari'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_characters' not in data.columns:
        data = get_num_characters(data, backbone=backbone)

    data = data.with_columns(
        (4.71 * (pl.col("n_characters") / pl.col("n_tokens")) + \
         0.5 * (pl.col("n_tokens") / pl.col("n_sentences")) - 21.43
         ).alias("ari"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("ari")

    return data

def get_smog(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the Simple Measure of Gobbledygook (SMOG) of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
                Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Simple Measure of
            Gobbledygook of the text data. The Simple Measure of
            Gobbledygook is stored in a new column named 'smog'.
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_polysyllables' not in data.columns:
        data = get_num_polysyllables(data, backbone=backbone)

    data = data.with_columns(
        (1.0430 * (30 * pl.col("n_polysyllables") /  \
                   pl.col("n_sentences"))**0.5 + 3.1291
         ).alias("smog"),
    )
    if data.filter(pl.col("n_sentences") == 0).shape[0] > 0:
        zero_token_warning_nan("smog")

    return data

def get_cli(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the Coleman-Liau Index (CLI) of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Coleman-Liau Index
            of the text data. The Coleman-Liau Index is stored
            in a new column named 'cli'.
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_characters' not in data.columns:
        data = get_num_characters(data, backbone=backbone)
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)

    data = data.with_columns(
        (0.0588 * (pl.col("n_characters") / pl.col("n_tokens") * 100) - \
         0.296 * (pl.col("n_sentences") / pl.col("n_tokens") * 100) - 15.8
         ).alias("cli"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("cli")

    return data

def get_gunning_fog(data: pl.DataFrame,
                    backbone: str = 'spacy',
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the Gunning Fog Index of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the Gunning Fog Index
            of the text data. The Gunning Fog Index is stored
            in a new column named 'gunning_fog'.
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_polysyllables' not in data.columns:
        data = get_num_polysyllables(data, backbone=backbone)
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)

    data = data.with_columns(
        (0.4 * ((pl.col("n_tokens") / pl.col("n_sentences")) + \
                100 * (pl.col("n_polysyllables") / pl.col("n_tokens")))
                ).alias("gunning_fog"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("gunning_fog")

    return data

def get_lix(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the LIX of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the LIX of the text data.
            The LIX is stored in a new column named 'lix'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_long_words' not in data.columns:
        data = get_num_long_words(data, backbone=backbone)

    data = data.with_columns(
        (pl.col("n_tokens") / pl.col("n_sentences") + \
         100 * pl.col("n_long_words") / pl.col("n_tokens")
         ).alias("lix"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("lix")

    return data

def get_rix(data: pl.DataFrame,
            backbone: str = 'spacy',
            **kwargs: dict[str, str],
            ) -> pl.DataFrame:
    """
    Calculates the RIX of a text.

    Args:
        data (pl.DataFrame): APolars DataFrame containing the text data.
        backbone (str):
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
            Not supported for Stanza backbone.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the RIX of the text data.
            The RIX is stored in a new column named 'rix'.
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'n_long_words' not in data.columns:
        data = get_num_long_words(data, backbone=backbone)

    data = data.with_columns(
        (pl.col("n_long_words") / pl.col("n_sentences")
         ).alias("rix"),
    )
    if data.filter(pl.col("n_tokens") == 0).shape[0] > 0:
        zero_token_warning_nan("rix")

    return data




"""
This module contains functions to calculate various surface-level features
from text data.

The surface-level features implemented in this module are:

- Raw Sequence Length:
    The number of characters in the text including whitespaces.
- Number of Tokens: The number of tokens in the text.
- Number of Sentences: The number of sentences in the text.
- Number of Tokens per Sentence: The average number of tokens per sentence.
- Number of Characters:
    The number of characters in the text excluding whitespaces.
- Number of Characters per Sentence:
    The average number of characters per sentence.
- Raw Length per Sentence:
    The average number of characters per sentence including whitespaces.
- Average Word Length: The average length of a word in the text.
- Number of Types: The number of unique tokens in the text.
- Number of Long Words: The number of words longer than a threshold.
- Number of Lemmas: The number of unique lemmas in the text.
- Token Frequencies: The frequency of each token in the text.
"""

from collections import Counter

import polars as pl

def get_raw_sequence_length(data: pl.DataFrame,
                            text_column: str = 'text',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the raw text length (number of characters) of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        text_column (str): The name of the column containing the text data.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the raw text length of the text
            data. The raw text length is stored in a new column named
            'raw_sequence_length'.
    """
    data = data.with_columns(
        pl.col(text_column).map_elements(lambda x: len(x),
                                         return_dtype=pl.UInt16
                                         ).alias("raw_sequence_length"),
    )
    
    return data

def get_num_tokens(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Calculates the sequence length (number of tokens) of a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the sequence length of the text
            data. The sequence length is stored in a new column named
            'n_tokens'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(x),
                                       return_dtype=pl.UInt16
                                       ).alias("n_tokens"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: x.num_tokens,
                                       return_dtype=pl.UInt16
                                       ).alias("n_tokens"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
    
    return data

def get_num_sentences(data: pl.DataFrame,
                      backbone: str = 'spacy',
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
    """
    Calculates the number of sentences in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of sentences in the
            text data. The number of sentences is stored in a new column
            named 'n_sentences'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(list(x.sents)),
                                         return_dtype=pl.UInt16
                                       ).alias("n_sentences"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(x.sentences),
                                       return_dtype=pl.UInt16
                                       ).alias("n_sentences"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
    
    return data

def get_num_tokens_per_sentence(data: pl.DataFrame,
                                backbone: str = 'spacy',
                                **kwargs: dict[str, str],
                                ) -> pl.DataFrame:
    """
    Calculates the average number of tokens per sentence in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average number of tokens per
            sentence in the text data. The average number of tokens per
            sentence is stored in a new column named 'tokens_per_sentence'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)

    data = data.with_columns(
        (pl.col("n_tokens") / pl.col("n_sentences")). \
            alias("tokens_per_sentence"),
    )

    return data

def get_num_characters(data: pl.DataFrame,
                    backbone: str = 'spacy',
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
        """
        Calculates the number of characters in a text.
        Only takes tokens into account in contrast to
        get_raw_sequence_length.

        Args:
            data (pl.DataFrame):
                A Polars DataFrame containing the text data.
            backbone (str): The NLP library used to process the text data.
                    Either 'spacy' or 'stanza'.
        
        Returns:
            data (pl.DataFrame):
                A Polars DataFrame containing the number of characters in 
                the text data. The number of characters is stored in a new
                column named 'n_characters'.
        """
        if backbone == 'spacy':
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: sum(
                    [len(token.text) for token in x]),
                    return_dtype=pl.UInt16
                    ).alias("n_characters"),
            )
        elif backbone == 'stanza':
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x: sum(
                    [len(token.text) for sent
                     in x.sentences for token in sent.tokens]),
                    return_dtype=pl.UInt16
                    ).alias("n_characters"),
            )
        else:
            raise ValueError(f"Unsupported backbone '{backbone}'. "
                             "Supported backbones are 'spacy' and 'stanza'.")
        
        return data

def get_chars_per_sentence(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the average number of characters per sentence in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average number of characters
            per sentence in the text data. The average number of
            characters per sentence is stored in a new column named
            'characters_per_sentence'.
    """
    if 'n_characters' not in data.columns:
        data = get_num_characters(data, backbone=backbone)
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)

    data = data.with_columns(
        (
            pl.col("n_characters") / pl.col("n_sentences")
        ).alias("characters_per_sentence"),
    )

    return data

def get_raw_length_per_sentence(data: pl.DataFrame,
                                backbone: str = 'spacy',
                                text_column: str = 'text',
                                **kwargs: dict[str, str],
                                ) -> pl.DataFrame:
    """
    Calculates the average number of characters per sentence in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average number of characters
            per sentence in the text data. The average number of
            characters per sentence is stored in a new column named
            'raw_length_per_sentence'.
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone=backbone)
    if 'raw_sequence_length' not in data.columns:
        data = get_raw_sequence_length(data, text_column=text_column)

    data = data.with_columns(
        (
            pl.col("raw_sequence_length") / pl.col("n_sentences")
        ).alias("raw_length_per_sentence"),
    )

    return data

def get_avg_word_length(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        text_column: str = 'text'
                        ,**kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the average word length in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the average word length in the
            text data.
            The average word length is stored in a new column named
            'avg_word_length'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'n_characters' not in data.columns:
        data = get_num_characters(data, text_column=text_column)

    data = data.with_columns(
        (
            pl.col("n_characters") / pl.col("n_tokens")
        ).alias("avg_word_length"),
    )

    return data

def get_num_types(data: pl.DataFrame,
                  backbone: str = 'spacy',
                  **kwargs: dict[str, str],
                 ) -> pl.DataFrame:
    """
    Calculates the number of types in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the number of types in the text 
            data.
            The number of types is stored in a new column named 'n_types'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp"). \
                map_elements(lambda x: len(
                    set([token.text for token in x])),
                             return_dtype=pl.UInt16
                             ).alias("n_types"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp"). \
                map_elements(lambda x: len(
                    set([token.text for sent
                         in x.sentences for token in sent.tokens])),
                         return_dtype=pl.UInt16
                         ).alias("n_types"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
        
    return data

def get_num_long_words(data: pl.DataFrame,
                       backbone: str = 'spacy',
                       threshold: int = 6,
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Calculates the number of long words in a text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        threshold (int):
            The minimum length of a word to be considered long.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of long words in the 
            text data.
            The number of long words is stored in a new column named 
            'n_long_words'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp"). \
                map_elements(lambda x: len(
                    [token for token in x if len(token.text) >= threshold]),
                             return_dtype=pl.UInt16
                             ).alias("n_long_words"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp"). \
                map_elements(lambda x: len(
                    [token for sent
                     in x.sentences for token in sent.tokens
                     if len(token.text) >= threshold]),
                             return_dtype=pl.UInt16
                             ).alias("n_long_words"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
        
    return data

def get_num_lemmas(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Calculates the number of unique lemmas in the text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the number of unique lemmas in
            the text data. The number of unique lemmas is stored in a new 
            column named 'n_lemmas'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(set(
                [token.lemma_ for token in x])),
                return_dtype=pl.UInt16
                ).alias("n_lemmas"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: len(set(
                [token.lemma for sent in x.sentences for token
                 in sent.words])),
                return_dtype=pl.UInt16
                ).alias("n_lemmas"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
        
    return data

def get_token_freqs(data: pl.DataFrame,
                    backbone: str = 'spacy',
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Calculates the frequency of each token in the text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the frequency of each token in
            the text data. The frequency of each token is stored in a new 
            column named 'token_freqs'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: dict(
                Counter([token.text for token in x])),
                return_dtype=pl.Object
                ).alias("token_freqs"),
        )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x: dict(
                Counter([token.text for sent in x.sentences for token
                         in sent.tokens])),
                return_dtype=pl.Object
                ).alias("token_freqs"),
        )
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")
    
    return data

def get_global_token_frequencies(data: pl.DataFrame,
                                 backbone: str = 'spacy',
                                 **kwargs: dict[str, str],
                                 ) -> dict[str, int]:
    """
    Calculates the global frequency of each token in the text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        token_freqs (dict[str, int]):
            A dictionary containing the frequency of each token in the text.
    """
    if backbone == 'spacy':
        token_freqs = dict(Counter([token.text for text in 
                                    data["nlp"].to_list() for 
                                    token in text]
        ))
    elif backbone == 'stanza':
        token_freqs = dict(Counter([token.text for text in
                                    data["nlp"].to_list() for
                                    sent in text.sentences for 
                                    token in sent.tokens]
        ))
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")

    return token_freqs

def get_global_lemma_frequencies(data: pl.DataFrame,
                                 backbone: str = 'spacy',
                                 **kwargs: dict[str, str],
                                 ) -> dict[str, int]:
    """
    Calculates the global frequency of each lemma in the text.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        lemma_freqs (dict[str, int]):
            A dictionary containing the frequency of each lemma in the text.
    """
    if backbone == 'spacy':
        lemma_freqs = dict(Counter([token.lemma_ for text in
                                    data["nlp"].to_list() for
                                    token in text]
        ))
    elif backbone == 'stanza':
        lemma_freqs = dict(Counter([token.lemma for text in
                                    data["nlp"].to_list() for
                                    sent in text.sentences for
                                    token in sent.tokens]
        ))
    else:
        raise ValueError(f"Unsupported backbone '{backbone}'. "
                         "Supported backbones are 'spacy' and 'stanza'.")

    return lemma_freqs




"""
This module contains functions to extract dependency features from the
input text. Dependency features are based on the dependency tree of the
text data. The dependency tree is a directed graph that represents the
syntactic structure of the text. Each node in the graph represents a
token in the text, and each edge represents a dependency relation between
two tokens. Dependency features can be used to capture the syntactic
complexity of the text data.

The following dependency features are implemented in this module:

- Tree width:
    The maximum number of nodes in the dependency tree of a token.
- Tree depth:
    The maximum distance of a token from the root of the dependency tree.
- Tree branching factor: 
    The average number of children of a token.
- Ramification factor: 
    The mean number of children per level in the dependency tree.
- Number of noun chunks:
    The number of noun chunks in the text.
- Frequency per dependency type:
    The frequency of each dependency type in the text.
"""
import numpy as np
import polars as pl

from .configs.dependency_config import (
    CLEARNLP_DEPENDENCIES_CONFIG,
    UNIVERSAL_DEPENDENCIES_CONFIG,
)

# ---------------------- Dependency Tree Features ---------------------- #

def get_tree_width(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Extracts the dependency tree width of the input text.
    Tree width is the maximum number of nodes in the dependency
    tree of a token.

    Args:
        data (pl.DataFrame): A polars dataframe containing the text data.
        backbone (str): The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            The input data with the dependency tree width stored in a new
            column named 'tree_width'.
    """
    def get_width(nlp):
        if len(nlp) == 0:  # empty input
            return 0
        return max([len(list(token.children)) for token in nlp])
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col('nlp').map_elements(
                lambda x: get_width(x),
            return_dtype=pl.UInt16
            ).alias('tree_width')
        )
    elif backbone == 'stanza':
        print('Dependency tree width extraction is not yet implemented'
              ' for Stanza.')
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")

    return data

def get_tree_depth(data: pl.DataFrame,
                   backbone: str = 'spacy',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Extracts the dependency tree depth of the input text.
    Tree depth is the maximum distance of a token from the root
    of the dependency tree.
    If the input text contains multiple sentences, the average
    tree depth is calculated.

    Args:
        data (pl.DataFrame): A polars dataframe containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            The input data with the dependency tree depth stored in a new
            column named 'tree_depth'.
    """
    def walk_tree(token, depth):
        if len(list(token.children)) == 0:
            return depth
        return max([walk_tree(child, depth + 1) for child in
                    token.children])
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col('nlp').map_elements(
                lambda x: np.mean([walk_tree(sent.root, 0) for sent in
                                   x.sents]) if len(list(x.sents)) > 0 else 0,
            return_dtype=pl.Float64
            ).alias('tree_depth')
        )
    elif backbone == 'stanza':
        print('Dependency tree depth extraction is not yet implemented '
              'for Stanza.')
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")

    return data

def get_tree_branching(data: pl.DataFrame,
                       backbone: str = 'spacy',
                       **kwargs: dict[str, str],
                       ) -> pl.DataFrame:
    """
    Extracts the dependency tree branching factor of the input text.
    The branching factor is the average number of children of a token.

    Args:
       data (pl.DataFrame): A polars dataframe containing the text data.
       backbone (str): The NLP library used to process the text data.
               Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            The input data with the dependency tree branching factor
            stored in a new column named 'tree_branching'.
    """
    def get_branching(nlp):
        if len(nlp) == 0:  # empty input
            return 0
        return sum([len(list(token.children)) for token in nlp]) / len(nlp)
    if backbone == 'spacy':
         data = data.with_columns(
               pl.col('nlp').map_elements(
                lambda x: get_branching(x),
                return_dtype=pl.Float32
               ).alias('tree_branching')
         )
    elif backbone == 'stanza':
         print('Dependency tree branching factor extraction is not yet'
               ' implemented for Stanza.')
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")

    return data

def get_ramification_factor(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Extracts the dependency tree ramification factor of the input text.
    The ramification factor is the mean number of children per level.

    Args:
        data (pl.DataFrame): A polars dataframe containing the text data.
        backbone (str): The NLP library used to process the text data.
                        Either 'spacy' or 'stanza'.
    
    Returns:
        data (pl.DataFrame):
            The input data with the dependency tree ramification factor
            stored in a new column named 'ramification_factor'.
    """
    def get_ramification(nlp):
        if len(nlp) == 0:  # empty input
            return 0
        levels = {}
        for token in nlp:
            if token.dep not in levels:
                levels[token.dep] = 0
            levels[token.dep] += len(list(token.children))
        return sum(levels.values()) / len(levels)
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col('nlp').map_elements(
                lambda x: get_ramification(x),
                return_dtype=pl.Float32
            ).alias('ramification_factor')
        )
    elif backbone == 'stanza':
        print('Dependency tree ramification factor extraction is not yet'
              ' implemented for Stanza.')
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")

    return data

def get_n_noun_chunks(data: pl.DataFrame,
                      backbone: str = 'spacy',
                      **kwargs: dict[str, str],
                      ) -> pl.DataFrame:
    """
    Extracts the number of noun chunks in the input text.

    Args:
        data (pl.DataFrame): A polars dataframe containing the text data.
        backbone (str): The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame): 
            The input data with the number of noun chunks stored in a new
            column named 'n_noun_chunks'.
    """
    def get_n_chunks(nlp):
        return len(list(nlp.noun_chunks))
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col('nlp').map_elements(
                lambda x: get_n_chunks(x),
                return_dtype=pl.UInt16
            ).alias('n_noun_chunks')
        )
    elif backbone == 'stanza':
        print('Number of noun chunks extraction is not yet implemented'
              ' for Stanza.')
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")

    return data

# ----------------- Dependency Relation/Type Features ------------------ #

def get_n_per_dependency_type(data: pl.DataFrame,
                              backbone: str = 'spacy',
                              language: str = 'en',
                              **kwargs: dict[str, str],
                              ) -> pl.DataFrame:
    """
    Extracts the frequency per dependency type in the input text.

    Args:
        data (pl.DataFrame): A polars dataframe containing the text data.
        backbone (str): The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
        language (str): The language of the text data. E.g. 'en', 'de'.
    
    Returns:
        data (pl.DataFrame):
            The input data with the frequency per dependency type stored
            in new columns named 'n_dependency_{dep}'.
    """
    if backbone == 'spacy':
        if language in ['en', 'de']:
            dependencies = CLEARNLP_DEPENDENCIES_CONFIG
        else:
            dependencies = UNIVERSAL_DEPENDENCIES_CONFIG

        for dep in dependencies:
            data = data.with_columns(
                pl.col('nlp').map_elements(
                    lambda x: len([token for token in x if
                                    token.dep_ == dep]),
                return_dtype=pl.UInt16
                ).alias(f'n_dependency_{dep}')
            )
    elif backbone == 'stanza':
        dependencies = UNIVERSAL_DEPENDENCIES_CONFIG
        for dep in dependencies:
            data = data.with_columns(
                pl.col('nlp').map_elements(
                    lambda x: len([token for sent in x.sentences for \
                                   token in sent.words if \
                                    token.deprel == dep]),
                return_dtype=pl.UInt16
                ).alias(f'n_dependency_{dep}')
            )
    else:
        raise ValueError(f"Unsupported NLP library: {backbone}. "
                         "Please use 'spacy' or 'stanza'.")
    
    return data

